from __future__ import annotations

import weakref
from functools import partialmethod
from typing import Any
from typing import TypeVar

from seabreeze.pyseabreeze.exceptions import SeaBreezeError
from seabreeze.pyseabreeze.types import PySeaBreezeProtocol

FT = TypeVar("FT", bound="SeaBreezeFeature")


class SeaBreezeFeature:
    identifier = "base_feature"

    _required_kwargs: tuple[str, ...] = ()
    _required_features: tuple[str, ...] = ()
    _required_protocol_cls = PySeaBreezeProtocol

    def __init__(
        self, protocol: PySeaBreezeProtocol, feature_id: int, **kwargs: Any
    ) -> None:
        """SeaBreezeFeature base class

        Parameters
        ----------
        protocol : Type[seabreeze.pyseabreeze.protocol.ProtocolInterface}
        feature_id : int
        """
        if self.identifier == "base_feature":
            raise SeaBreezeError(
                "Don't instantiate SeaBreezeFeature directly. Use derived feature classes."
            )
        assert set(self._required_kwargs) == set(kwargs), "{} vs {}".format(
            str(set(self._required_kwargs)), str(set(kwargs))
        )
        # check protocol support
        if not isinstance(protocol, self._required_protocol_cls):
            raise SeaBreezeError("FeatureError: Protocol not supported by feature")
        self.protocol: PySeaBreezeProtocol = weakref.proxy(protocol)
        self.feature_id = feature_id

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.feature_id}>"

    @classmethod
    def get_feature_class_registry(cls) -> dict[str, type[SeaBreezeFeature]]:
        # noinspection PyUnresolvedReferences
        return {
            feature_class.identifier: feature_class
            for feature_class in SeaBreezeFeature.__subclasses__()
        }

    @classmethod
    def supports_protocol(cls, protocol: PySeaBreezeProtocol) -> bool:
        return isinstance(protocol, cls._required_protocol_cls)

    @classmethod
    def specialize(cls: type[FT], model_name: str, **kwargs: Any) -> type[FT]:
        assert set(kwargs) == set(cls._required_kwargs)
        specialized_class = type(
            f"{cls.__name__}{model_name}",
            (cls,),
            {"__init__": partialmethod(cls.__init__, **kwargs)},
        )
        return specialized_class



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeMulticastFeature(SeaBreezeFeature):
    identifier = "multicast"

    def get_multicast_enable_state(self, interface_index: int) -> bool:
        raise NotImplementedError("implement in derived class")

    def set_multicast_enable_state(
        self, interface_index: int, enable_state: bool
    ) -> None:
        raise NotImplementedError("implement in derived class")



from __future__ import annotations

import weakref
from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Generic
from typing import Iterable
from typing import TypeVar


class PySeaBreezeProtocol(ABC):
    def __init__(
        self,
        transport: PySeaBreezeTransport[Any],
    ) -> None:
        if not isinstance(transport, PySeaBreezeTransport):
            raise TypeError("transport does not implement read and write methods")
        self.transport: PySeaBreezeTransport[Any] = weakref.proxy(transport)

    @abstractmethod
    def send(
        self,
        msg_type: int,
        payload: tuple[int | str | float, ...] | str | int | float = (),
        timeout_ms: int | None = None,
        **kwargs: Any,
    ) -> int:
        ...

    @abstractmethod
    def receive(
        self,
        size: int | None = None,
        timeout_ms: int | None = None,
        **kwargs: Any,
    ) -> bytes:
        ...

    @abstractmethod
    def query(
        self,
        msg_type: int,
        payload: tuple[int | str | float, ...] | str | int | float = (),
        timeout_ms: int | None = None,
        **kwargs: Any,
    ) -> bytes:
        ...


DT = TypeVar("DT")


class PySeaBreezeTransport(ABC, Generic[DT]):

    _required_init_kwargs: tuple[str, ...] = ()

    @abstractmethod
    def open_device(self, device: DT) -> None:
        """open a seabreeze device

        Parameters
        ----------
        device : seabreeze.pyseabreeze.devices.SeaBreezeDevice
        """
        ...

    @property
    @abstractmethod
    def is_open(self) -> bool:
        """return if device is opened

        Returns
        -------
        bool
        """
        return False

    @abstractmethod
    def close_device(self) -> None:
        """close the seabreeze device"""
        ...

    @abstractmethod
    def write(self, data: bytes, timeout_ms: int | None = None, **kwargs: Any) -> int:
        """write data to the device"""
        ...

    @abstractmethod
    def read(
        self,
        size: int | None = None,
        timeout_ms: int | None = None,
        **kwargs: Any,
    ) -> bytes:
        """read data from the

        Returns
        -------
        bytes
        """
        ...

    @property
    @abstractmethod
    def default_timeout_ms(self) -> int:
        ...

    @property
    @abstractmethod
    def protocol(self) -> PySeaBreezeProtocol:
        ...

    @classmethod
    @abstractmethod
    def list_devices(cls) -> Iterable[DT]:
        ...

    @classmethod
    @abstractmethod
    def register_model(cls, model_name: str, **kwargs: str) -> None:
        ...

    @classmethod
    @abstractmethod
    def supported_model(cls, device: DT) -> str | None:
        """return if the device supports the transport or vice versa

        Returns
        -------
        model_name : ``
        """
        return None

    @classmethod
    @abstractmethod
    def specialize(
        cls, model_name: str, **kwargs: str
    ) -> type[PySeaBreezeTransport[Any]]:
        ...



from __future__ import annotations

from typing import TYPE_CHECKING

from seabreeze.pyseabreeze.features._base import SeaBreezeFeature

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeIrradCalFeature(SeaBreezeFeature):
    identifier = "irrad_cal"

    def has_collection_area(self) -> bool:
        raise NotImplementedError("implement in derived class")

    def read_collection_area(self) -> float:
        raise NotImplementedError("implement in derived class")

    def write_collection_area(self, area: float) -> None:
        raise NotImplementedError("implement in derived class")

    def read_calibration(self) -> NDArray[np.int_]:
        raise NotImplementedError("implement in derived class")

    def write_calibration(self, calibration_array: NDArray[np.int_]) -> None:
        raise NotImplementedError("implement in derived class")



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeStrobeLampFeature(SeaBreezeFeature):
    identifier = "strobe_lamp"

    def enable_lamp(self, state: bool) -> None:
        raise NotImplementedError("implement in derived class")



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeI2CMasterFeature(SeaBreezeFeature):
    identifier = "i2c_master"

    def get_number_of_buses(self) -> int:
        raise NotImplementedError("implement in derived class")

    def read_bus(
        self, bus_index: int, slave_address: int, buffer_length: int = 1024
    ) -> bytes:
        raise NotImplementedError("implement in derived class")

    def write_bus(self, bus_index: int, slave_address: int, data: bytes) -> None:
        raise NotImplementedError("implement in derived class")



print("import: 'seabreeze'")
import seabreeze

print("import: 'seabreeze.types'")
import seabreeze.types

print("import: 'seabreeze.cseabreeze'")
import seabreeze.cseabreeze

print("import: 'seabreeze.pyseabreeze'")
import seabreeze.pyseabreeze




"""This is a wrapper for the c-implementation of the seabreeze-library

"""
# noinspection PyProtectedMember
from seabreeze.cseabreeze._wrapper import SeaBreezeAcquisitionDelayFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeAPI
from seabreeze.cseabreeze._wrapper import SeaBreezeContinuousStrobeFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeDataBufferFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeDevice
from seabreeze.cseabreeze._wrapper import SeaBreezeDHCPServerFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeEEPROMFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeError
from seabreeze.cseabreeze._wrapper import SeaBreezeEthernetConfigurationFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeFastBufferFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeGPIOFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeI2CMasterFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeIntrospectionFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeIPv4Feature
from seabreeze.cseabreeze._wrapper import SeaBreezeIrradCalFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeLightSourceFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeMulticastFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeNetworkConfigurationFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeNonlinearityCoefficientsFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeNumFeaturesError
from seabreeze.cseabreeze._wrapper import SeaBreezeOpticalBenchFeature
from seabreeze.cseabreeze._wrapper import SeaBreezePixelBinningFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeRawUSBBusAccessFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeRevisionFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeShutterFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeSpectrometerFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeSpectrumProcessingFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeStrayLightCoefficientsFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeStrobeLampFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeTemperatureFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeThermoElectricFeature
from seabreeze.cseabreeze._wrapper import SeaBreezeWifiConfigurationFeature

__all__ = [
    "_backend_",
    "SeaBreezeAcquisitionDelayFeature",
    "SeaBreezeAPI",
    "SeaBreezeContinuousStrobeFeature",
    "SeaBreezeDataBufferFeature",
    "SeaBreezeDevice",
    "SeaBreezeDHCPServerFeature",
    "SeaBreezeEEPROMFeature",
    "SeaBreezeError",
    "SeaBreezeEthernetConfigurationFeature",
    "SeaBreezeFastBufferFeature",
    "SeaBreezeFeature",
    "SeaBreezeGPIOFeature",
    "SeaBreezeI2CMasterFeature",
    "SeaBreezeIntrospectionFeature",
    "SeaBreezeIPv4Feature",
    "SeaBreezeIrradCalFeature",
    "SeaBreezeLightSourceFeature",
    "SeaBreezeMulticastFeature",
    "SeaBreezeNetworkConfigurationFeature",
    "SeaBreezeNonlinearityCoefficientsFeature",
    "SeaBreezeNumFeaturesError",
    "SeaBreezeOpticalBenchFeature",
    "SeaBreezePixelBinningFeature",
    "SeaBreezeRawUSBBusAccessFeature",
    "SeaBreezeRevisionFeature",
    "SeaBreezeShutterFeature",
    "SeaBreezeSpectrometerFeature",
    "SeaBreezeSpectrumProcessingFeature",
    "SeaBreezeStrayLightCoefficientsFeature",
    "SeaBreezeStrobeLampFeature",
    "SeaBreezeTemperatureFeature",
    "SeaBreezeThermoElectricFeature",
    "SeaBreezeWifiConfigurationFeature",
]

_backend_ = "cseabreeze"



import struct
from unittest import mock

import pytest

from seabreeze.pyseabreeze.types import PySeaBreezeTransport


@pytest.fixture
def obp_message(pyseabreeze):
    """create a valid OBP message"""
    from seabreeze.pyseabreeze.protocol import OBPProtocol

    _o = OBPProtocol.OBP
    yield struct.pack(
        _o.HEADER_FMT + "0s" + _o.FOOTER_FMT,
        _o.HEADER_START_BYTES,
        _o.HEADER_PROTOCOL_VERSION,
        0,
        _o.NO_ERROR,
        0x00000100,
        0,
        _o.RESERVED,
        _o.CHECKSUM_TYPE_NONE,
        1,
        b"c",
        20,
        b"",
        _o.NO_CHECKSUM,
        _o.FOOTER,
    )


@pytest.fixture
def mock_transport(obp_message):
    mock_transport = mock.Mock(spec=PySeaBreezeTransport)
    mock_transport.read.return_value = obp_message
    yield mock_transport


def test_pyseabreeze_protocol_messages(mock_transport):
    """test the parsing code for the OBP protocol implementation of pyseabreeze"""
    from seabreeze.pyseabreeze.protocol import OBPProtocol

    obp = OBPProtocol(mock_transport)
    obp.send(0x00000100)



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeEthernetConfigurationFeature(SeaBreezeFeature):
    identifier = "ethernet_configuration"

    def get_mac_address(self, interface_index: int) -> str:
        raise NotImplementedError("implement in derived class")

    def set_mac_address(self, interface_index: int, mac_address: int) -> None:
        raise NotImplementedError("implement in derived class")

    def get_gbe_enable_status(self, interface_index: int) -> bool:
        raise NotImplementedError("implement in derived class")

    def set_gbe_enable_status(self, interface_index: int, enable_state: bool) -> None:
        raise NotImplementedError("implement in derived class")



"""seabreeze.types

common types used to teach mypy about seabreeze
"""
from __future__ import annotations

import sys
from typing import TYPE_CHECKING
from typing import Any
from typing import TypeVar

if sys.version_info >= (3, 8):
    from typing import Literal
    from typing import Protocol
    from typing import TypedDict
    from typing import runtime_checkable
else:
    from typing_extensions import Literal
    from typing_extensions import Protocol
    from typing_extensions import TypedDict
    from typing_extensions import runtime_checkable

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray

__all__ = [
    "SeaBreezeAPI",
    "SeaBreezeBackend",
    "SeaBreezeDevice",
    "SeaBreezeFeatureAccessor",
    "SeaBreezeFeatureDict",
]


@runtime_checkable
class SeaBreezeDevice(Protocol):
    serial_number: str
    model: str  # fixme
    is_open: bool

    features: SeaBreezeFeatureDict
    f: SeaBreezeFeatureAccessor

    def open(self) -> None:
        ...

    def close(self) -> None:
        ...


class SeaBreezeAPI(Protocol):
    """SeaBreeze API interface"""

    def __init__(self, initialize: bool = True, **_kwargs: Any) -> None:
        raise TypeError(f"Type {type(self).__name__} cannot be instantiated")

    def initialize(self) -> None:
        ...

    def shutdown(self) -> None:
        ...

    def add_rs232_device_location(
        self, device_type: str, bus_path: str, baudrate: int
    ) -> None:
        ...

    def add_ipv4_device_location(
        self, device_type: str, ip_address: str, port: int
    ) -> None:
        ...

    def list_devices(self) -> list[SeaBreezeDevice]:
        ...

    @staticmethod
    def supported_models() -> list[str]:
        ...


FT = TypeVar("FT", bound="SeaBreezeFeature")


class SeaBreezeFeature(Protocol):
    identifier: str

    def __repr__(self) -> str:
        ...

    @classmethod
    def get_feature_class_registry(cls) -> dict[str, FT]:
        ...

    @classmethod
    def supports_protocol(cls, protocol: Any) -> bool:
        ...

    @classmethod
    def specialize(
        cls: type[SeaBreezeFeature], model_name: str, **kwargs: str
    ) -> type[FT]:
        ...


class SeaBreezeAcquisitionDelayFeature(SeaBreezeFeature):
    def set_delay_microseconds(self, delay_usec: int) -> None:
        ...

    def get_delay_microseconds(self) -> int:
        ...

    delay_microseconds: int

    def get_delay_increment_microseconds(self) -> int:
        ...

    def get_maximum_delay_microseconds(self) -> int:
        ...

    def get_minimum_delay_microseconds(self) -> int:
        ...


class SeaBreezeContinuousStrobeFeature(SeaBreezeFeature):
    def set_enable(self, strobe_enable: bool) -> None:
        ...

    def set_period_micros(self, period_micros: int) -> None:
        ...


class SeaBreezeDataBufferFeature(SeaBreezeFeature):
    def clear(self) -> None:
        ...

    def remove_oldest_spectra(self, number_of_spectra: int) -> None:
        ...

    def get_number_of_elements(self) -> int:
        ...

    def get_buffer_capacity(self) -> int:
        ...

    def set_buffer_capacity(self, capacity: int) -> None:
        ...

    def get_buffer_capacity_maximum(self) -> int:
        ...

    def get_buffer_capacity_minimum(self) -> int:
        ...


class SeaBreezeDHCPServerFeature(SeaBreezeFeature):
    def get_address(self, interface_index: int) -> str:
        ...

    def set_address(self, interface_index: int, server_address: str) -> None:
        ...

    def get_enable_state(self, interface_index: int) -> bool:
        ...

    def set_enable_state(self, interface_index: int, enable_state: bool) -> None:
        ...


class SeaBreezeEEPROMFeature(SeaBreezeFeature):
    def eeprom_read_slot(self, slot_number: int, strip_zero_bytes: bool = False) -> str:
        ...


class SeaBreezeEthernetConfigurationFeature(SeaBreezeFeature):
    def get_mac_address(self, interface_index: int) -> str:
        ...

    def set_mac_address(self, interface_index: int, mac_address: str) -> None:
        ...

    def get_gbe_enable_status(self, interface_index: int) -> bool:
        ...

    def set_gbe_enable_status(self, interface_index: int, enable_state: bool) -> None:
        ...


class SeaBreezeFastBufferFeature(SeaBreezeFeature):
    def get_buffering_enable(self) -> bool:
        ...

    def set_buffering_enable(self, is_enabled: bool) -> None:
        ...

    def get_consecutive_sample_count(self) -> int:
        ...

    def set_consecutive_sample_count(self, consecutive_sample_count: int) -> None:
        ...


class SeaBreezeGPIOFeature(SeaBreezeFeature):
    modes: dict[str, int]

    def get_number_of_gpio_pins(self) -> int:
        ...

    def get_gpio_output_enable_vector(self) -> int:
        ...

    def set_gpio_output_enable_vector(
        self, output_enable_vector: int, bit_mask: int
    ) -> None:
        ...

    def get_gpio_value_vector(self) -> int:
        ...

    def set_gpio_value_vector(self, value_vector: int, bit_mask: int) -> None:
        ...

    def get_number_of_egpio_pins(self) -> int:
        ...

    def get_egpio_available_modes(self, pin_number: int) -> int:
        ...

    def get_egpio_current_mode(self, pin_number: int) -> int:
        ...

    def set_egpio_mode(self, pin_number: int, mode: int, value: float = 0.0) -> None:
        ...

    def get_egpio_output_vector_vector(self) -> int:
        ...

    def set_egpio_output_vector(self, output_vector: int, bit_mask: int) -> None:
        ...

    def get_egpio_value(self, pin_number: int) -> int:
        ...

    def set_egpio_value(self, pin_number: int, value: int) -> None:
        ...


class SeaBreezeI2CMasterFeature(SeaBreezeFeature):
    def get_number_of_buses(self) -> int:
        ...

    def read_bus(
        self, bus_index: int, slave_address: int, buffer_length: int = 1024
    ) -> bytes:
        ...

    def write_bus(self, bus_index: int, slave_address: int, data: bytes) -> None:
        ...


class SeaBreezeIntrospectionFeature(SeaBreezeFeature):
    def number_of_pixels(self) -> int:
        ...

    def get_active_pixel_ranges(self) -> tuple[tuple[int, int], ...]:
        ...

    def get_optical_dark_pixel_ranges(self) -> tuple[tuple[int, int], ...]:
        ...

    def get_electric_dark_pixel_ranges(self) -> tuple[tuple[int, int], ...]:
        ...


class SeaBreezeIPv4Feature(SeaBreezeFeature):
    def get_dhcp_enable_state(self, interface_index: int) -> bool:
        ...

    def set_dhcp_enable_state(self, interface_index: int, is_enabled: bool) -> None:
        ...

    def get_number_of_ipv4_addresses(self, interface_index: int) -> int:
        ...

    def get_ipv4_address(self, interface_index: int, address_index: int) -> str:
        ...

    def get_default_gateway(self, interface_index: int) -> str:
        ...

    def set_default_gateway(
        self, interface_index: int, default_gateway_address: str
    ) -> None:
        ...

    def add_ipv4_address(self, interface_index: int, ipv4_address: str) -> None:
        ...

    def delete_ipv4_address(self, interface_index: int, address_index: int) -> None:
        ...


class SeaBreezeIrradCalFeature(SeaBreezeFeature):
    def has_collection_area(self) -> bool:
        ...

    def read_collection_area(self) -> float:
        ...

    def write_collection_area(self, area: float) -> None:
        ...

    def read_calibration(self) -> NDArray[np.int_]:
        ...

    def write_calibration(self, calibration_array: NDArray[np.int_]) -> None:
        ...


class SeaBreezeLightSourceFeature(SeaBreezeFeature):
    def get_count(self) -> int:
        ...

    def has_enable(self, light_source_index: int) -> bool:
        ...

    def is_enabled(self, light_source_index: int) -> bool:
        ...

    def set_enable(self, light_source_index: int, enable: bool) -> None:
        ...

    def has_variable_intensity(self, light_source_index: int) -> bool:
        ...

    def get_intensity(self, light_source_index: int) -> int:
        ...

    def set_intensity(self, light_source_index: int, intensity: int) -> None:
        ...


class SeaBreezeMulticastFeature(SeaBreezeFeature):
    def get_multicast_enable_state(self, interface_index: int) -> bool:
        ...

    def set_multicast_enable_state(
        self, interface_index: int, enable_state: bool
    ) -> None:
        ...


class SeaBreezeNetworkConfigurationFeature(SeaBreezeFeature):
    def get_number_of_network_interfaces(self) -> int:
        ...

    def get_network_interface_connection_type(self, interface_index: int) -> int:
        ...

    def get_network_interface_enable_state(self, interface_index: int) -> bool:
        ...

    def set_network_interface_enable_state(
        self, interface_index: int, enable_state: bool
    ) -> None:
        ...

    def run_network_interface_self_test(self, interface_index: int) -> bool:
        ...

    def save_network_interface_connection_settings(self, interface_index: int) -> None:
        ...


class SeaBreezeNonlinearityCoefficientsFeature(SeaBreezeFeature):
    def get_nonlinearity_coefficients(self) -> list[float]:
        ...


class SeaBreezeOpticalBenchFeature(SeaBreezeFeature):
    def get_fiber_diameter_microns(self) -> int:
        ...

    def get_slit_width_microns(self) -> int:
        ...

    def get_id(self) -> str:
        ...

    def get_serial_number(self) -> str:
        ...

    def get_coating(self) -> str:
        ...

    def get_filter(self) -> str:
        ...

    def get_grating(self) -> str:
        ...


class SeaBreezePixelBinningFeature(SeaBreezeFeature):
    def set_binning_factor(self, factor: int) -> None:
        ...

    def get_binning_factor(self) -> int:
        ...

    def get_max_binning_factor(self) -> int:
        ...

    binning_factor: int
    max_binning_factor: int

    def set_default_binning_factor(self, factor: int) -> None:
        ...

    def get_default_binning_factor(self) -> int:
        ...

    default_binning_factor: int


class SeaBreezeRawUSBBusAccessFeature(SeaBreezeFeature):
    def raw_usb_write(
        self, data: bytes, endpoint: Literal["primary_out", "secondary_out"]
    ) -> None:
        ...

    def raw_usb_read(
        self,
        endpoint: Literal["primary_in", "secondary_in", "secondary_in2"],
        buffer_length: int | None = None,
    ) -> bytes:
        ...


class SeaBreezeRevisionFeature(SeaBreezeFeature):
    def hardware_revision(self) -> str:
        ...

    def revision_firmware_get(self) -> str:
        ...


class SeaBreezeShutterFeature(SeaBreezeFeature):
    def set_shutter_open(self, state: bool) -> None:
        ...


class SeaBreezeSpectrumProcessingFeature(SeaBreezeFeature):
    def get_boxcar_width(self) -> int:
        ...

    def set_boxcar_width(self, boxcar_width: int) -> None:
        ...

    boxcar_width: int

    def get_scans_to_average(self) -> int:
        ...

    def set_scans_to_average(self, scans_to_average: int) -> None:
        ...

    scans_to_average: int


class SeaBreezeStrayLightCoefficientsFeature(SeaBreezeFeature):
    def get_stray_light_coefficients(self) -> list[float]:
        ...


class SeaBreezeStrobeLampFeature(SeaBreezeFeature):
    def enable_lamp(self, state: bool) -> None:
        ...


class SeaBreezeTemperatureFeature(SeaBreezeFeature):
    def count_temperatures(self) -> int:
        ...

    def read_temperature(self, index: int) -> float:
        ...

    def temperature_get_all(self) -> list[float]:
        ...


class SeaBreezeThermoElectricFeature(SeaBreezeFeature):
    def read_temperature_degrees_celsius(self) -> float:
        ...

    def set_temperature_setpoint_degrees_celsius(self, temperature: float) -> None:
        ...

    def enable_tec(self, state: bool) -> None:
        ...


class SeaBreezeWifiConfigurationFeature(SeaBreezeFeature):
    def get_wifi_mode(self, interface_index: int) -> int:
        ...

    def set_wifi_mode(self, interface_index: int, wifi_mode: int) -> None:
        ...

    def get_wifi_security_type(self, interface_index: int) -> int:
        ...

    def set_wifi_security_type(self, interface_index: int, security_type: int) -> None:
        ...

    def get_wifi_ssid(self, interface_index: int) -> str:
        ...

    def set_wifi_ssid(self, interface_index: int, ssid: str) -> None:
        ...

    def set_wifi_pass_phrase(self, interface_index: int, pass_phrase: str) -> None:
        ...


class SeaBreezeSpectrometerFeature(SeaBreezeFeature):
    def set_trigger_mode(self, mode: int) -> None:
        ...

    def set_integration_time_micros(self, integration_time_micros: int) -> None:
        ...

    def get_integration_time_micros_limits(self) -> tuple[int, int]:
        ...

    def get_maximum_intensity(self) -> float:
        ...

    def get_electric_dark_pixel_indices(self) -> list[int]:
        ...

    _spectrum_length: int

    def get_wavelengths(self) -> NDArray[np.float_]:
        ...

    def get_intensities(self) -> NDArray[np.float_]:
        ...

    def _get_spectrum_raw(self) -> bytes:
        ...

    def get_fast_buffer_spectrum(self) -> Any:
        ...  # fixme


class SeaBreezeFeatureAccessor(Protocol):
    acquisition_delay: SeaBreezeAcquisitionDelayFeature
    continuous_strobe: SeaBreezeContinuousStrobeFeature
    data_buffer: SeaBreezeDataBufferFeature
    dhcp_server: SeaBreezeDHCPServerFeature
    eeprom: SeaBreezeEEPROMFeature
    ethernet_configuration: SeaBreezeEthernetConfigurationFeature
    fast_buffer: SeaBreezeFastBufferFeature
    gpio: SeaBreezeGPIOFeature
    i2c_master: SeaBreezeI2CMasterFeature
    introspection: SeaBreezeIntrospectionFeature
    ipv4: SeaBreezeIPv4Feature
    irrad_cal: SeaBreezeIrradCalFeature
    light_source: SeaBreezeLightSourceFeature
    multicast: SeaBreezeMulticastFeature
    network_configuration: SeaBreezeNetworkConfigurationFeature
    nonlinearity_coefficients: SeaBreezeNonlinearityCoefficientsFeature
    optical_bench: SeaBreezeOpticalBenchFeature
    pixel_binning: SeaBreezePixelBinningFeature
    raw_usb_bus_access: SeaBreezeRawUSBBusAccessFeature
    revision: SeaBreezeRevisionFeature
    shutter: SeaBreezeShutterFeature
    spectrometer: SeaBreezeSpectrometerFeature
    spectrum_processing: SeaBreezeSpectrumProcessingFeature
    stray_light_coefficients: SeaBreezeStrayLightCoefficientsFeature
    strobe_lamp: SeaBreezeStrobeLampFeature
    temperature: SeaBreezeTemperatureFeature
    thermo_electric: SeaBreezeThermoElectricFeature
    wifi_configuration: SeaBreezeWifiConfigurationFeature


class SeaBreezeFeatureDict(TypedDict):
    acquisition_delay: list[SeaBreezeAcquisitionDelayFeature]
    continuous_strobe: list[SeaBreezeContinuousStrobeFeature]
    data_buffer: list[SeaBreezeDataBufferFeature]
    dhcp_server: list[SeaBreezeDHCPServerFeature]
    eeprom: list[SeaBreezeEEPROMFeature]
    ethernet_configuration: list[SeaBreezeEthernetConfigurationFeature]
    fast_buffer: list[SeaBreezeFastBufferFeature]
    gpio: list[SeaBreezeGPIOFeature]
    i2c_master: list[SeaBreezeI2CMasterFeature]
    introspection: list[SeaBreezeIntrospectionFeature]
    ipv4: list[SeaBreezeIPv4Feature]
    irrad_cal: list[SeaBreezeIrradCalFeature]
    light_source: list[SeaBreezeLightSourceFeature]
    multicast: list[SeaBreezeMulticastFeature]
    network_configuration: list[SeaBreezeNetworkConfigurationFeature]
    nonlinearity_coefficients: list[SeaBreezeNonlinearityCoefficientsFeature]
    optical_bench: list[SeaBreezeOpticalBenchFeature]
    pixel_binning: list[SeaBreezePixelBinningFeature]
    raw_usb_bus_access: list[SeaBreezeRawUSBBusAccessFeature]
    revision: list[SeaBreezeRevisionFeature]
    shutter: list[SeaBreezeShutterFeature]
    spectrometer: list[SeaBreezeSpectrometerFeature]
    spectrum_processing: list[SeaBreezeSpectrumProcessingFeature]
    stray_light_coefficients: list[SeaBreezeStrayLightCoefficientsFeature]
    strobe_lamp: list[SeaBreezeStrobeLampFeature]
    temperature: list[SeaBreezeTemperatureFeature]
    thermo_electric: list[SeaBreezeThermoElectricFeature]
    wifi_configuration: list[SeaBreezeWifiConfigurationFeature]


class SeaBreezeBackend(Protocol):
    """SeaBreeze Backend interface"""

    _backend_: str
    _api_kwargs: dict[str, Any]

    SeaBreezeAPI: type[SeaBreezeAPI]
    SeaBreezeDevice: type[SeaBreezeDevice]
    SeaBreezeError: type[Exception]
    SeaBreezeFeature: type[SeaBreezeFeature]

    SeaBreezeAcquisitionDelayFeature: type[SeaBreezeAcquisitionDelayFeature]
    SeaBreezeContinuousStrobeFeature: type[SeaBreezeContinuousStrobeFeature]
    SeaBreezeDataBufferFeature: type[SeaBreezeDataBufferFeature]
    SeaBreezeDHCPServerFeature: type[SeaBreezeDHCPServerFeature]
    SeaBreezeEEPROMFeature: type[SeaBreezeEEPROMFeature]
    SeaBreezeEthernetConfigurationFeature: type[SeaBreezeEthernetConfigurationFeature]
    SeaBreezeFastBufferFeature: type[SeaBreezeFastBufferFeature]
    SeaBreezeGPIOFeature: type[SeaBreezeGPIOFeature]
    SeaBreezeI2CMasterFeature: type[SeaBreezeI2CMasterFeature]
    SeaBreezeIntrospectionFeature: type[SeaBreezeIntrospectionFeature]
    SeaBreezeIPv4Feature: type[SeaBreezeIPv4Feature]
    SeaBreezeIrradCalFeature: type[SeaBreezeIrradCalFeature]
    SeaBreezeLightSourceFeature: type[SeaBreezeLightSourceFeature]
    SeaBreezeMulticastFeature: type[SeaBreezeMulticastFeature]
    SeaBreezeNetworkConfigurationFeature: type[SeaBreezeNetworkConfigurationFeature]
    SeaBreezeNonlinearityCoefficientsFeature: type[
        SeaBreezeNonlinearityCoefficientsFeature
    ]
    SeaBreezeNumFeaturesError: type[Exception]
    SeaBreezeOpticalBenchFeature: type[SeaBreezeOpticalBenchFeature]
    SeaBreezePixelBinningFeature: type[SeaBreezePixelBinningFeature]
    SeaBreezeRawUSBBusAccessFeature: type[SeaBreezeRawUSBBusAccessFeature]
    SeaBreezeRevisionFeature: type[SeaBreezeRevisionFeature]
    SeaBreezeShutterFeature: type[SeaBreezeShutterFeature]
    SeaBreezeSpectrometerFeature: type[SeaBreezeSpectrometerFeature]
    SeaBreezeSpectrumProcessingFeature: type[SeaBreezeSpectrumProcessingFeature]
    SeaBreezeStrayLightCoefficientsFeature: type[SeaBreezeStrayLightCoefficientsFeature]
    SeaBreezeStrobeLampFeature: type[SeaBreezeStrobeLampFeature]
    SeaBreezeTemperatureFeature: type[SeaBreezeTemperatureFeature]
    SeaBreezeThermoElectricFeature: type[SeaBreezeThermoElectricFeature]
    SeaBreezeWifiConfigurationFeature: type[SeaBreezeWifiConfigurationFeature]



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeRevisionFeature(SeaBreezeFeature):
    identifier = "revision"

    def hardware_revision(self) -> str:
        raise NotImplementedError("implement in derived class")

    def revision_firmware_get(self) -> str:
        raise NotImplementedError("implement in derived class")



"""Implementation of the Protocol layer

ocean optics spectrometers use two different communication protocols:
- one is called 'OBP protocol'  # OceanBinaryProtocol
- the other one 'OOI protocol'  # ??? OceanOpticsInterface ??? maybe

"""
from __future__ import annotations

import functools
import hashlib
import struct
import time
import warnings
from typing import Any

from seabreeze.pyseabreeze.exceptions import SeaBreezeError
from seabreeze.pyseabreeze.types import PySeaBreezeProtocol
from seabreeze.pyseabreeze.types import PySeaBreezeTransport


class OOIProtocol(PySeaBreezeProtocol):

    msgs = {
        code: functools.partial(struct.Struct(msg).pack, code)
        for code, msg in {
            0x01: "<B",  # OP_INITIALIZE
            0x02: "<BI",  # OP_ITIME
            0x03: "<BH",  # set Strobe/Lamp enable Line
            0x05: "<BB",  # OP_GETINFO
            0x09: "<B",  # OP_REQUESTSPEC
            0x0A: "<BH",  # OP_SETTRIGMODE
            0x6A: "<BBH",  # OP_WRITE_REGISTER
            0x6B: "<BB",  # OP_READ_REGISTER
            0x71: "<BBB",  # OP_TECENABLE_QE
            0x72: "<B",  # OP_READTEC_QE
            0x73: "<Bh",  # OP_TECSETTEMP_QE
            0xFE: "<B",  # OP_USBMODE
        }.items()
    }  # add more here if you implement new features

    def __init__(self, transport: PySeaBreezeTransport[Any]) -> None:
        super().__init__(transport)
        # initialize the spectrometer
        self.send(0x01)
        time.sleep(0.1)  # wait shortly after init command

    def send(
        self,
        msg_type: int,
        payload: tuple[float | int | str, ...] | str | int | float = (),
        timeout_ms: int | None = None,
        **kwargs: str | int | None,
    ) -> int:
        """send a ooi message to the spectrometer

        Parameters
        ----------
        msg_type : int
            a message type as defined in `OOIProtocol.msgs`
        payload : optional
            dependent on `msg_type`. a singe value or a tuple of multiple values
        timeout_ms : int, optional
            the timeout after which the transport layer should error.
            `None` means no timeout (default)
        **kwargs :
            ignored and only present to provide compatible caller interfaces

        Returns
        -------
        bytes_written : int
            the number of bytes sent
        """
        if kwargs:
            warnings.warn(f"kwargs provided but ignored: {kwargs}")
        payload = payload if isinstance(payload, (tuple, list)) else (payload,)
        data = self.msgs[msg_type](*payload)
        return self.transport.write(data, timeout_ms=timeout_ms)

    def receive(
        self,
        size: int | None = None,
        timeout_ms: int | None = None,
        mode: str | None = None,
        **kwargs: Any,
    ) -> bytes:
        """receive data from the spectrometer

        Parameters
        ----------
        size:
            number of bytes to receive. if `None` (default) uses the
            default size as specified in the transport layer.
        timeout_ms:
            the timeout after which the transport layer should error.
            `None` means no timeout (default)
        mode:
            transport layers can support different modes
            (i.e. {'low_speed', 'high_speed', 'high_speed_alt'} in the usb case)
        kwargs:
            ignored and only present to provide compatible caller interfaces

        Returns
        -------
        data:
            data returned from the spectrometer
        """
        if kwargs:
            warnings.warn(f"kwargs provided but ignored: {kwargs}")
        return self.transport.read(
            size=size, timeout_ms=timeout_ms, mode=mode, **kwargs
        )

    def query(
        self,
        msg_type: int,
        payload: tuple[int | str | float, ...] | str | int | float = (),
        timeout_ms: int | None = None,
        size: int | None = None,
        mode: str | None = None,
        **kwargs: str | int | None,
    ) -> bytes:
        """convenience method combining send and receive

        Parameters
        ----------
        msg_type:
            a message type as defined in `OOIProtocol.msgs`
        payload:
            the payload to be sent. Can be a singe value or a tuple, dependent
            `msg_type`.
        size:
            number of bytes to receive. if `None` (default) uses the
            default size as specified in the transport layer.
        timeout_ms:
            the timeout after which the transport layer should error.
            `None` means no timeout (default)
        mode:
            transport layers can support different modes
            (i.e. {'low_speed', 'high_speed', 'high_speed_alt'} in the usb case)
        kwargs:
            ignored and only present to provide compatible caller interfaces

        Returns
        -------
        data:
            data returned from the spectrometer
        """
        self.send(msg_type, payload, timeout_ms=timeout_ms)
        return self.receive(size=size, timeout_ms=timeout_ms, mode=mode, **kwargs)


class OBPProtocol(PySeaBreezeProtocol):

    msgs = {
        code: struct.Struct(msg).pack
        for code, msg in {
            0x00000100: "",  # GET_SERIAL
            0x00000101: "",  # GET_SERIAL_LENGTH
            0x00100928: "",  # GET_BUF_SPEC32_META
            0x00101000: "",  # GET_RAW_SPECTRUM_NOW_HDX
            0x00101100: "",  # GET_RAW_SPECTRUM_NOW
            0x00110010: "<L",  # SET_ITIME_USEC
            0x00110110: "<B",  # SET_TRIG_MODE
            0x00180100: "",  # GET_WL_COEFF_COUNT
            0x00180101: "<B",  # GET_WL_COEFF
            0x00181100: "",  # GET_NL_COEFF_COUNT
            0x00181101: "<B",  # GET_NL_COEFF
            0x00420004: "",  # GET_TE_TEMPERATURE
            0x00420010: "<B",  # SET_TE_ENABLE
            0x00420011: "<f",  # SET_TE_SETPOINT
        }.items()
    }  # add more here if you implement new features

    class OBP:
        """All relevant constants are stored here"""

        HEADER_START_BYTES = 0xC0C1
        HEADER_PROTOCOL_VERSION = 0x1100  # XXX: seems to be the newest protocol version

        FLAG_RESPONSE_TO_REQUEST = 0x0001
        FLAG_ACK = 0x0002
        FLAG_REQUEST_ACK = 0x0004
        FLAG_NACK = 0x0008
        FLAG_HW_EXCEPTION = 0x0010
        FLAG_PROTOCOL_DEPRECATED = 0x0020

        ERROR_CODES = {
            0: "Success (no detectable errors)",
            1: "Invalid/unsupported protocol",
            2: "Unknown message type",
            3: "Bad checksum",
            4: "Message too large",
            5: "Payload length does not match message type",
            6: "Payload data invalid",
            7: "Device not ready for given message type",
            8: "Unknown checksum type",
            9: "Device reset unexpectedly",
            10: "Too many buses (Commands have come from too many bus interfaces)",
            11: "Out of memory. Failed to allocate enough space to complete request.",
            12: "Command is valid, but desired information does not exist.",
            13: "Int Device Error. May be unrecoverable.",
            100: "Could not decrypt properly",
            101: "Firmware layout invalid",
            102: "Data packet was wrong size",
            103: "hardware revision not compatible with firmware",
            104: "Existing flash map not compatible with firmware",
            255: "Operation/Response Deferred. Will take some time. Do not ACK or NACK yet.",
        }

        NO_ERROR = 0x0000
        RESERVED = b""
        CHECKSUM_TYPE_NONE = 0x00
        CHECKSUM_TYPE_MD5 = 0x01
        NO_CHECKSUM = b""
        FOOTER = 0xC2C3C4C5  # the datasheet specifies it in this order...

        HEADER_FMT = (
            "<H"  # start_bytes
            "H"  # protocol_version
            "H"  # flags
            "H"  # error number
            "L"  # message type
            "L"  # regarding
            "6s"  # reserved
            "B"  # checksum type
            "B"  # immediate length
            "16s"  # immediate data
            "L"  # bytes remaining
        )

        FOOTER_FMT = "16s" "L"  # checksum  # footer

    # noinspection DuplicatedCode
    def send(
        self,
        msg_type: int,
        payload: tuple[int | str | float, ...] | str | int | float = (),
        timeout_ms: int | None = None,
        request_ack: bool = True,
        **kwargs: Any,
    ) -> int:
        """send a obp message to the spectrometer

        Parameters
        ----------
        msg_type : int
            a message type as defined in `OBPProtocol.msgs`
        payload : optional
            dependent on `msg_type`. Can be a single value or a tuple of multiple values
        timeout_ms : int, optional
            the timeout after which the transport layer should error.
            `None` means no timeout (default)
        request_ack : bool, default `True`
            request an ack for the sent command from the spectrometer.
        **kwargs :
            ignored and only present to provide compatible caller interfaces

        Returns
        -------
        bytes_written : int
            the number of bytes sent
        """
        if kwargs:
            warnings.warn(f"kwargs provided but ignored: {kwargs}")
        payload = payload if isinstance(payload, (tuple, list)) else (payload,)
        data = self.msgs[msg_type](*payload)

        # Constructing message and querying usb.
        message = self._construct_outgoing_message(
            msg_type, data, request_ack=request_ack
        )  # TODO
        bytes_written = self.transport.write(
            message
        )  # ? assert bytes_written == len(message)

        if not request_ack:
            return bytes_written

        response = self.transport.read(timeout_ms=timeout_ms)
        try:
            remaining_bytes, checksum_type = self._check_incoming_message_header(
                response[:44]
            )
        except SeaBreezeError:
            return 0
        # ? assert remaining_bytes == 20
        checksum = self._check_incoming_message_footer(response[-20:])

        if (
            checksum_type == self.OBP.CHECKSUM_TYPE_MD5
            and checksum != hashlib.md5(response[:-20]).digest()
        ):
            warnings.warn(
                "WARNING OBP: The checksums differ, but we ignore this for now."
            )
        return bytes_written

    def receive(
        self,
        size: int | None = None,
        timeout_ms: int | None = None,
        **kwargs: Any,
    ) -> bytes:
        """receive data from the spectrometer

        Parameters
        ----------
        size : int, optional
            number of bytes to receive. if `None` (default) uses the
            default size as specified in the transport layer.
        timeout_ms : int, optional
            the timeout after which the transport layer should error.
            `None` means no timeout (default)
        kwargs :
            ignored and only present to provide compatible caller interfaces

        Returns
        -------
        data : str
            data returned from the spectrometer
        """
        response = self.transport.read(size=64, timeout_ms=timeout_ms)
        try:
            remaining_bytes, checksum_type = self._check_incoming_message_header(
                response[:44]
            )
        except SeaBreezeError:
            # empty buffer if error raised
            self.transport.read(size=None, timeout_ms=500)
            raise
        length_payload_footer = remaining_bytes

        # we already received some data
        remaining_bytes -= len(response[44:])
        if remaining_bytes > 0:
            response += self.transport.read(size=remaining_bytes, timeout_ms=timeout_ms)

        if length_payload_footer != len(response[44:]):
            raise SeaBreezeError(
                "remaining packet length mismatch: {:d} != {:d}".format(
                    remaining_bytes, len(response[44:])
                )
            )

        checksum = self._check_incoming_message_footer(response[-20:])
        if (
            checksum_type == self.OBP.CHECKSUM_TYPE_MD5
            and checksum != hashlib.md5(response[:-20]).digest()
        ):
            warnings.warn(
                "WARNING OBP: The checksums differ, but we ignore this for now."
            )

        return self._extract_message_data(response)

    def query(
        self,
        msg_type: int,
        payload: tuple[int | str | float, ...] | str | int | float = (),
        timeout_ms: int | None = None,
        size: int | None = None,
        **kwargs: Any,
    ) -> bytes:
        """convenience method combining send and receive

        Parameters
        ----------
        msg_type : int
            a message type as defined in `OBPProtocol.msgs`
        payload :
            the payload to be sent. Can be a single value or a tuple, dependent
            `msg_type`.
        size : int, optional
            number of bytes to receive. if `None` (default) uses the
            default size as specified in the transport layer.
        timeout_ms : int, optional
            the timeout after which the transport layer should error.
            `None` means no timeout (default)
        kwargs :
            ignored and only present to provide compatible caller interfaces

        Returns
        -------
        data : str
            data returned from the spectrometer
        """
        self.send(msg_type, payload, request_ack=False)
        return self.receive(timeout_ms=timeout_ms)

    def _construct_outgoing_message(
        self,
        msg_type: int,
        payload_string: bytes,
        request_ack: bool = False,
        regarding: int | None = None,
    ) -> bytes:
        """construct an outgoing OBP message

        Parameters
        ----------
        msg_type:
            the obp message type, a 4 byte integer
        payload_string:
            a compiled payload_string
        request_ack:
            request an ack for the sent command from the spectrometer.
        regarding:
            see ocean optics obp protocol documentation

        Returns
        -------
        message:
            compiled message
        """
        if request_ack is True:
            flags = self.OBP.FLAG_REQUEST_ACK
        else:
            flags = 0

        if regarding is None:
            regarding = 0

        if len(payload_string) <= 16:
            payload_string_fmt = "0s"
            immediate_length = len(payload_string)
            immediate_data = payload_string
            payload_string = b""
            bytes_remaining = 20  # Checksum + footer
        else:
            payload_string_fmt = "%ds" % len(payload_string)
            immediate_length = 0
            immediate_data = b""
            bytes_remaining = 20 + len(payload_string)

        FMT = self.OBP.HEADER_FMT + payload_string_fmt + self.OBP.FOOTER_FMT

        msg = struct.pack(
            FMT,
            self.OBP.HEADER_START_BYTES,
            self.OBP.HEADER_PROTOCOL_VERSION,
            flags,
            self.OBP.NO_ERROR,
            msg_type,
            regarding,
            self.OBP.RESERVED,
            self.OBP.CHECKSUM_TYPE_NONE,
            immediate_length,
            immediate_data,
            bytes_remaining,
            payload_string,
            self.OBP.NO_CHECKSUM,
            self.OBP.FOOTER,
        )
        return msg

    def _check_incoming_message_header(self, header: bytes) -> tuple[int, int]:
        """check the incoming message header

        Parameters
        ----------
        header:
            a obp header of length 44

        Returns
        -------
        bytes_and_checksum_type:
            bytes_remaining after the header (returns 20 for a 64 byte message)
            checksum_type only supports self.OBP.CHECKSUM_TYPE_MD5 for now
        """
        if len(header) != 44:
            raise SeaBreezeError(
                "header has wrong length! len(header): %d" % len(header)
            )

        data = struct.unpack(self.OBP.HEADER_FMT, header)

        if data[0] != self.OBP.HEADER_START_BYTES:
            raise SeaBreezeError('Header start_bytes wrong: "%d"' % data[0])
        if data[1] != self.OBP.HEADER_PROTOCOL_VERSION:
            raise SeaBreezeError("Header protocol version wrong: %d" % data[1])

        flags = data[2]
        if flags == 0:
            pass
        if flags & self.OBP.FLAG_RESPONSE_TO_REQUEST:
            pass  # TODO: propagate?
        if flags & self.OBP.FLAG_ACK:
            pass  # TODO: propagate?
        if flags & self.OBP.FLAG_REQUEST_ACK:
            pass  # TODO: only the host should be able to set this?
        if (flags & self.OBP.FLAG_NACK) or (flags & self.OBP.FLAG_HW_EXCEPTION):
            error = data[3]
            if error != 0:  # != SUCCESS
                raise SeaBreezeError(self.OBP.ERROR_CODES[error])
            else:
                pass  # TODO: should we do something here?
        if flags & self.OBP.FLAG_PROTOCOL_DEPRECATED:
            raise SeaBreezeError("Protocol deprecated?!?")

        # msg_type = data[4]
        # regarding = data[5]

        checksum_type = data[7]  # TODO: implement checksums.
        if checksum_type not in [
            self.OBP.CHECKSUM_TYPE_NONE,
            self.OBP.CHECKSUM_TYPE_MD5,
        ]:
            raise SeaBreezeError('the checksum type is unknown: "%d"' % checksum_type)

        # immediate_length = data[8]
        # immediate_data = data[9]
        bytes_remaining = data[10]

        return bytes_remaining, checksum_type

    def _check_incoming_message_footer(self, footer: bytes) -> bytes:
        """check the incoming message header

        Parameters
        ----------
        footer:
            a obp footer of length 20

        Returns
        -------
        checksum:
            the 16 byte checksum of the message
        """
        assert len(footer) == 20, "footer has wrong length! len(footer): %d" % len(
            footer
        )

        data = struct.unpack("<" + self.OBP.FOOTER_FMT, footer)

        checksum: bytes = data[0]
        assert data[1] == self.OBP.FOOTER, (
            "the device returned a wrong footer: %d" % data[1]
        )

        return checksum

    def _extract_message_data(self, msg: bytes) -> bytes:
        """extract the payload data from a obp message

        Parameters
        ----------
        msg:
            a obp message

        Returns
        -------
        data:
            the payload contained in the message
        """
        payload_length = len(msg) - 44 - 20  # - HeaderLength - FooterLength
        if not (payload_length >= 0):
            raise SeaBreezeError("Received message < 64 bytes: %d" % payload_length)
        payload_fmt = "%ds" % payload_length
        FMT = self.OBP.HEADER_FMT + payload_fmt + self.OBP.FOOTER_FMT

        data = struct.unpack(FMT, msg)

        msg_type = data[4]

        immediate_length: int = data[8]
        immediate_data: bytes = data[9]
        payload: bytes = data[11]

        if (immediate_length > 0) and len(payload) > 0:
            raise SeaBreezeError("Got immediate AND payload data? cmd: '%d'" % msg_type)
        elif immediate_length > 0:
            return immediate_data[:immediate_length]
        elif payload_length > 0:
            return payload
        else:
            return b""



"""device classes for pyseabreeze

Author: Andreas Poehlmann
"""
from __future__ import annotations

import enum
import itertools
from collections import defaultdict
from typing import Any
from typing import Iterable
from typing import Tuple
from typing import TypeVar

from seabreeze.pyseabreeze import features as sbf
from seabreeze.pyseabreeze.exceptions import SeaBreezeError
from seabreeze.pyseabreeze.features import SeaBreezeFeature
from seabreeze.pyseabreeze.protocol import OBPProtocol
from seabreeze.pyseabreeze.protocol import OOIProtocol
from seabreeze.pyseabreeze.transport import USBTransport
from seabreeze.pyseabreeze.types import PySeaBreezeTransport
from seabreeze.types import SeaBreezeFeatureAccessor

# class registry for all spectrometer models

_model_class_registry: dict[str, type[SeaBreezeDevice]] = {}


class _SeaBreezeDeviceMeta(type):
    """metaclass for pyseabreeze devices"""

    def __new__(
        mcs: type[_SeaBreezeDeviceMeta],
        name: str,
        bases: tuple[Any],
        attr_dict: dict[str, Any],
    ) -> _SeaBreezeDeviceMeta:
        # This part of the metaclass magic is very opaque. It could be avoided by moving all
        # of this logic to custom __init__ methods of the Spectrometer classes, or to factory
        # functions. But I am using python-seabreeze as a playground to experiment with
        # different ways of implementing extendable interfaces, so please forgive these design
        # decisions ^^"
        #
        # I also like the concise way you can define the spectrometer functionality this way.
        # It allows you to get a quick overview of the functionality implemented in pyseabreeze,
        # without having to read through a lot of code. And I think that it's easy to understand
        # what a spectrometer can do, by just looking at the spectrometer class definition, even
        # if you don't understand the details of the implementation.
        #
        # Because it's quite opaque the code below tries hard to enforce strict typing in all
        # defined spectrometer classes to minimize the amount of errors you can make when adding
        # a new spectrometer to pyseabreeze.
        #
        if name != "SeaBreezeDevice":
            # This runs for all subclasses of SeaBreezeDevice, so for all defined spectrometers
            # What might be unintuitive to the user is, that all defined attributes in the
            # spectrometer classes are only used to configure the subclass and are NOT directly
            # available in the instances later. (look at how attr_dict is modified below).
            #
            if "model_name" not in attr_dict:
                raise AttributeError(f"'model_name' not provided for class '{name}'")
            model_name = attr_dict.pop("model_name")
            if not isinstance(model_name, str):
                raise TypeError(f"{name}.model_name not a str")

            new_attr_dict: dict[str, Any] = {
                "_model_name": model_name,
            }

            if any(base is SeaBreezeDevice for base in bases):
                # gather the transport classes defined on the class
                transport_classes = mcs._extract_transform_classes(
                    model_name, class_name=name, attr_dict=attr_dict
                )
                new_attr_dict["_transport_classes"] = transport_classes
            # gather the feature classes defined on the class
            feature_classes = mcs._extract_feature_classes(
                model_name, class_name=name, attr_dict=attr_dict
            )
            new_attr_dict["_feature_classes"] = feature_classes

            try:
                # move classmethod to new class for dynamic model class subsitution
                k = "_substitute_compatible_subclass"
                new_attr_dict[k] = attr_dict[k]
            except KeyError:
                pass

            if any(not attr.startswith("_") for attr in attr_dict):
                raise ValueError(
                    "can't define extra attrs on spectrometer classes: {}".format(
                        ", ".join(
                            attr for attr in attr_dict if not attr.startswith("_")
                        )
                    )
                )
            else:
                attr_dict = new_attr_dict

        return super().__new__(mcs, name, bases, attr_dict)

    def __init__(cls, name: str, bases: tuple[Any], attr_dict: dict[str, Any]) -> None:

        if name != "SeaBreezeDevice":
            # > model name
            model_name = getattr(cls, "_model_name")
            assert isinstance(model_name, str), "model name not a str"
            _model_class_registry[model_name] = cls  # type: ignore

        super().__init__(name, bases, attr_dict)

    @staticmethod
    def _extract_transform_classes(
        model_name: str, class_name: str, attr_dict: dict[str, Any]
    ) -> tuple[type[PySeaBreezeTransport[Any]], ...]:

        visited_attrs = set()
        transport_classes = []
        try:
            supported_transport_classes = attr_dict.pop("transport")
        except KeyError:
            raise AttributeError(f"{class_name}.transport not provided")
        if (
            not isinstance(supported_transport_classes, tuple)
            or not supported_transport_classes
        ):
            raise TypeError(f"{class_name}.transport not a tuple of len > 0")

        for idx, transport_cls in enumerate(supported_transport_classes):
            # for each supported transport of the spectrometer, gather the configuration from
            # the spectrometer class and specialize the transport_cls with the provided settings.
            #
            if not issubclass(transport_cls, PySeaBreezeTransport):
                raise TypeError(
                    "{}.transport[{:d}] '{}' does not derive from TransportInterface".format(
                        class_name, idx, transport_cls.__name__
                    )
                )
            # noinspection PyProtectedMember
            kwargs = transport_cls._required_init_kwargs
            transport_init_kwargs = {}
            for kw in kwargs:
                if kw not in attr_dict:
                    raise AttributeError(
                        "{}.{} not provided for class but '{}' requires it.".format(
                            class_name, kw, transport_cls.__name__
                        )
                    )
                transport_init_kwargs[kw] = attr_dict[kw]
            visited_attrs.update(kwargs)
            # specialize the transport class with the spectrometer's custom config
            specialized_transport_cls = transport_cls.specialize(
                model_name, **transport_init_kwargs
            )
            transport_classes.append(specialized_transport_cls)

        for attr in visited_attrs:
            del attr_dict[attr]

        return tuple(transport_classes)

    @staticmethod
    def _extract_feature_classes(
        model_name: str, class_name: str, attr_dict: dict[str, Any]
    ) -> dict[str, list[type[SeaBreezeFeature]]]:

        visited_attrs = set()
        feature_classes: defaultdict[str, list[type[SeaBreezeFeature]]] = defaultdict(
            list
        )
        try:
            supported_feature_classes = attr_dict.pop("feature_classes")
        except KeyError:
            raise AttributeError(f"{class_name}.feature_classes not provided")
        if (
            not isinstance(supported_feature_classes, tuple)
            or not supported_feature_classes
        ):
            raise TypeError(f"{class_name}.feature_classes not a tuple of len > 0")
        for idx, feature_cls in enumerate(supported_feature_classes):
            # for each supported feature of the spectrometer, gather the configuration
            # from the spectrometer class and subclass the feature_cls with the provided
            # settings. Also check if requirements are fulfilled, i.e. if the feature depends
            # on other features or on specific protocols
            #
            if not issubclass(feature_cls, SeaBreezeFeature):
                raise TypeError(
                    "{}.feature_classes[{:d}] '{}' does not derive from SeaBreezeFeature".format(
                        class_name, idx, feature_cls.__name__
                    )
                )
            # noinspection PyProtectedMember
            required = set(feature_cls._required_features)
            if not required.issubset(feature_classes):
                raise KeyError(
                    "{}.feature_classes[{:d}] '{}' requires '{}'. To fix, re-order or add.".format(
                        class_name,
                        idx,
                        feature_cls.__name__,
                        ", ".join(required - set(feature_classes)),
                    )
                )
            # noinspection PyProtectedMember
            kwargs = feature_cls._required_kwargs
            feature_attrs = {}
            for kw in kwargs:
                if kw not in attr_dict:
                    raise AttributeError(
                        "{}.{} not provided for class but '{}' requires it.".format(
                            class_name, kw, feature_cls.__name__
                        )
                    )
                feature_attrs[kw] = attr_dict[kw]
            visited_attrs.update(kwargs)
            # specialize the feature class with the spectrometer's custom config
            specialized_feature_cls = feature_cls.specialize(
                model_name, **feature_attrs
            )
            feature_classes[feature_cls.identifier].append(specialized_feature_cls)

        for attr in visited_attrs:
            del attr_dict[attr]

        return feature_classes


class EndPointMap:
    """internal endpoint map for spectrometer classes"""

    def __init__(
        self,
        ep_out: int | None = None,
        lowspeed_in: int | None = None,
        highspeed_in: int | None = None,
        highspeed_in2: int | None = None,
    ) -> None:
        self.primary_out = self.ep_out = ep_out
        self.primary_in = self.lowspeed_in = lowspeed_in
        self.secondary_out = ep_out
        self.secondary_in = self.highspeed_in = highspeed_in
        self.secondary_in2 = self.highspeed_in2 = highspeed_in2


class DarkPixelIndices(Tuple[int, ...]):
    """internal dark pixel range class"""

    def __new__(
        cls: type[DarkPixelIndices], indices: Iterable[int]
    ) -> DarkPixelIndices:
        """dark pixel indices

        Parameters
        ----------
        indices : iterable
            index of electric dark pixel
        """
        return super().__new__(DarkPixelIndices, sorted(set(indices)))  # type: ignore

    @classmethod
    def from_ranges(cls, *ranges: tuple[int, int]) -> DarkPixelIndices:
        """return dark pixes indices from ranges

        Parameters
        ----------
        *ranges : (`int`, `int`)
            ranges of electric dark pixels
        """
        dp = itertools.chain(*(range(low, high) for (low, high) in ranges))
        # noinspection PyArgumentList
        return cls(dp)


class TriggerMode(enum.IntEnum):
    """internal trigger modes enum"""

    NORMAL = 0x00
    SOFTWARE = 0x01
    LEVEL = 0x01
    SYNCHRONIZATION = 0x02
    HARDWARE = 0x03
    EDGE = 0x03
    SINGLE_SHOT = 0x04
    SELF_NORMAL = 0x80
    SELF_SOFTWARE = 0x81
    SELF_SYNCHRONIZATION = 0x82
    SELF_HARDWARE = 0x83
    DISABLED = 0xFF
    OBP_NORMAL = 0x00
    OBP_EXTERNAL = 0x01
    OBP_INTERNAL = 0x02
    OBP_EDGE = 0x01
    OBP_LEVEL = 0x03

    @classmethod
    def supported(cls, *mode_strings: str) -> set[TriggerMode]:
        return {getattr(cls, mode_string) for mode_string in mode_strings}


DT = TypeVar("DT", bound="SeaBreezeDevice")


class SeaBreezeDevice(metaclass=_SeaBreezeDeviceMeta):

    # internal attribute
    _model_name = None
    _serial_number = "?"
    _cached_features: dict[str, list[SeaBreezeFeature]] | None = None
    _transport_classes: tuple[type[PySeaBreezeTransport[Any]], ...]
    _feature_classes: dict[str, list[type[SeaBreezeFeature]]]

    def __new__(cls: type[DT], raw_device: Any = None) -> SeaBreezeDevice:
        if raw_device is None:
            raise SeaBreezeError(
                "Don't instantiate SeaBreezeDevice directly. Use `SeabreezeAPI.list_devices()`."
            )
        for transport in {USBTransport}:
            supported_model = transport.supported_model(raw_device)
            if supported_model is not None:
                break
        else:
            raise TypeError("No transport supports device.")
        specialized_cls = _model_class_registry[supported_model]
        return super().__new__(specialized_cls)

    def __init__(self, raw_device: Any = None) -> None:
        if raw_device is None:
            raise SeaBreezeError(
                "Don't instantiate SeaBreezeDevice directly. Use `SeabreezeAPI.list_devices()`."
            )
        self._raw_device = raw_device
        for transport in self._transport_classes:
            if transport.supported_model(self._raw_device) is not None:
                self._transport: PySeaBreezeTransport[Any] = transport()
                break
        else:
            raise TypeError("No transport supports device.")
        try:
            # sneakily switch in the correct subclass in case it's needed
            self.__class__ = self.__class__._substitute_compatible_subclass(
                self._transport
            )
        except AttributeError:
            pass
        try:
            self._serial_number = self.get_serial_number()
        except SeaBreezeError:
            pass

    @property
    def model(self) -> str:
        if self._model_name is None:
            raise RuntimeError("model can't be None")
        return self._model_name

    @property
    def serial_number(self) -> str:
        return self._serial_number

    @classmethod
    def _substitute_compatible_subclass(
        cls: type[DT], transport: PySeaBreezeTransport[Any]
    ) -> type[DT]:
        return cls

    def __repr__(self) -> str:
        return f"<SeaBreezeDevice {self.model}:{self.serial_number}>"

    def open(self) -> None:
        """open the spectrometer usb connection

        Returns
        -------
        None
        """
        self._transport.open_device(self._raw_device)
        # substitute subclass if needed
        self.__class__ = self.__class__._substitute_compatible_subclass(self._transport)
        # cache features
        self._cached_features = None
        _ = self.features
        # get serial
        self._serial_number = self.get_serial_number()

    def close(self) -> None:
        """close the spectrometer usb connection

        Returns
        -------
        None
        """
        if self.is_open:
            self._transport.close_device()

    @property
    def is_open(self) -> bool:
        """returns if the spectrometer device usb connection is opened

        Returns
        -------
        bool
        """
        return self._transport.is_open

    def get_serial_number(self) -> str:
        """return the serial number string of the spectrometer

        Returns
        -------
        serial_number: str
        """
        try:
            protocol = self._transport.protocol
        except RuntimeError:
            raise SeaBreezeError("device not open")

        if isinstance(protocol, OOIProtocol):
            # The serial is stored in slot 0
            return self.f.eeprom.eeprom_read_slot(0)

        elif isinstance(protocol, OBPProtocol):
            serial_len = ord(protocol.query(0x00000101))
            serial_str = protocol.query(0x00000100)[:serial_len]
            return serial_str.decode("utf8")

        else:
            raise NotImplementedError(
                f"No serial number for protocol class {type(protocol).__name__}"
            )

    @property
    def features(self) -> dict[str, list[SeaBreezeFeature]]:
        """return a dictionary of all supported features

        this returns a dictionary with all supported Features of the spectrometer

        Returns
        -------
        features : `dict` [`str`, `seabreeze.cseabreeze.SeaBreezeFeature`]
        """
        if not self._cached_features:
            protocol = self._transport.protocol
            self._cached_features = {}
            for identifier in sbf.SeaBreezeFeature.get_feature_class_registry():
                f_list = self._cached_features.setdefault(identifier, [])
                for feature_cls in self._feature_classes[identifier]:
                    assert (
                        issubclass(feature_cls, sbf.SeaBreezeFeature)
                        and identifier == feature_cls.identifier
                    )
                    if not feature_cls.supports_protocol(protocol):
                        continue
                    f_list.append(feature_cls(self._transport.protocol, len(f_list)))
        return self._cached_features

    @property
    def f(self) -> SeaBreezeFeatureAccessor:
        """convenience access to features via attributes

        this allows you to access a feature like this::

            # via .features
            device.features['spectrometer'][0].get_intensities()
            # via .f
            device.f.spectrometer.get_intensities()

        """

        class FeatureAccessHandler:
            def __init__(self, feature_dict: dict[str, Any]) -> None:
                for identifier, features in feature_dict.items():
                    setattr(
                        self, identifier, features[0] if features else None
                    )  # TODO: raise FeatureNotAvailable?

        accessor: SeaBreezeFeatureAccessor = FeatureAccessHandler(self.features)  # type: ignore
        return accessor


# SPECTROMETER DEFINITIONS
# ========================
#
class USB2000PLUS(SeaBreezeDevice):

    model_name = "USB2000PLUS"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x101E
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((6, 21))  # as in seabreeze-3.0.9
    integration_time_min = 1000
    integration_time_max = 655350000
    integration_time_base = 1
    spectrum_num_pixel = 2048
    spectrum_raw_length = (2048 * 2) + 1
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE", "SINGLE_SHOT"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureUSB2000PLUS,
        sbf.nonlinearity.NonlinearityCoefficientsEEPromFeatureOOI,
        sbf.continuousstrobe.SeaBreezeContinuousStrobeFeatureOOI,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )

    @classmethod
    def _substitute_compatible_subclass(
        cls: type[DT], transport: PySeaBreezeTransport[Any]
    ) -> type[DT]:
        """return the correct subclass of the usb2000plus like model"""
        try:
            protocol = transport.protocol
        except RuntimeError:
            raise AttributeError("transport not opened")

        # noinspection PyUnresolvedReferences,PyProtectedMember
        from seabreeze.pyseabreeze.features.fpga import _FPGARegisterFeatureOOI

        fpga = _FPGARegisterFeatureOOI(protocol)
        if fpga.get_firmware_version()[0] >= 3:
            return FLAMES  # type: ignore
        else:
            return USB2000PLUS  # type: ignore


class FLAMES(USB2000PLUS):

    model_name = "FLAMES"

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((6, 21))  # as in seabreeze-3.0.9
    integration_time_min = 1000
    integration_time_max = 655350000
    integration_time_base = 1
    spectrum_num_pixel = 2048
    spectrum_raw_length = (2048 * 2) + 1
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE", "SINGLE_SHOT"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureUSB2000PLUS,
        sbf.nonlinearity.NonlinearityCoefficientsEEPromFeatureOOI,
        sbf.continuousstrobe.SeaBreezeContinuousStrobeFeatureOOI,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class USB2000(SeaBreezeDevice):

    model_name = "USB2000"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1002
    usb_endpoint_map = EndPointMap(ep_out=0x02, lowspeed_in=0x87, highspeed_in=0x82)
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((2, 24))
    integration_time_min = 3000
    integration_time_max = 655350000
    integration_time_base = 1000
    spectrum_num_pixel = 2048
    spectrum_raw_length = (2048 * 2) + 1
    spectrum_max_value = 4095
    trigger_modes = TriggerMode.supported("NORMAL", "SOFTWARE", "HARDWARE")

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureUSB2000,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class HR2000(SeaBreezeDevice):

    model_name = "HR2000"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x100A
    usb_endpoint_map = EndPointMap(ep_out=0x02, lowspeed_in=0x87, highspeed_in=0x82)
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((2, 24))
    integration_time_min = 3000
    integration_time_max = 655350000
    integration_time_base = 1000
    spectrum_num_pixel = 2048
    spectrum_raw_length = (2048 * 2) + 1
    spectrum_max_value = 4095
    trigger_modes = TriggerMode.supported("NORMAL", "SOFTWARE", "HARDWARE")

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureHR2000,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class HR4000(SeaBreezeDevice):

    model_name = "HR4000"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1012
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((2, 13))
    integration_time_min = 10
    integration_time_max = 655350000
    integration_time_base = 1
    spectrum_num_pixel = 3840
    spectrum_raw_length = (3840 * 2) + 1
    spectrum_max_value = 16383
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureHR4000,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class HR2000PLUS(SeaBreezeDevice):

    model_name = "HR2000PLUS"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1016
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((2, 24))
    integration_time_min = 1000
    integration_time_max = 655350000
    integration_time_base = 1
    spectrum_num_pixel = 2048
    spectrum_raw_length = (2048 * 2) + 1
    spectrum_max_value = 16383
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureHR2000PLUS,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class USB650(SeaBreezeDevice):

    model_name = "USB650"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1014
    usb_endpoint_map = EndPointMap(ep_out=0x02, lowspeed_in=0x87, highspeed_in=0x82)
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges()
    integration_time_min = 3000
    integration_time_max = 655350000
    integration_time_base = 1000
    spectrum_num_pixel = 2048
    spectrum_raw_length = (2048 * 2) + 1
    spectrum_max_value = 4095
    trigger_modes = TriggerMode.supported("NORMAL", "SOFTWARE", "HARDWARE")

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureUSB650,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
        sbf.continuousstrobe.SeaBreezeContinuousStrobeFeatureOOI,
    )


class QE65000(SeaBreezeDevice):

    model_name = "QE65000"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1018
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges(
        (0, 4), (1040, 1044)
    )  # as in seabreeze-3.0.5
    integration_time_min = 8000
    integration_time_max = 1600000000
    integration_time_base = 1000
    spectrum_num_pixel = 1280
    spectrum_raw_length = (1024 + 256) * 2 + 1
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported("NORMAL", "SOFTWARE", "HARDWARE")

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureQE65000,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class USB4000(SeaBreezeDevice):

    model_name = "USB4000"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1022
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = (
        DarkPixelIndices.from_ranges((5, 16)),
    )  # as in seabreeze-3.0.9
    integration_time_min = 10
    integration_time_max = 65535000
    integration_time_base = 1
    spectrum_num_pixel = 3840
    spectrum_raw_length = (3840 * 2) + 1
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureUSB4000,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class NIRQUEST512(SeaBreezeDevice):

    model_name = "NIRQUEST512"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1026
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = (DarkPixelIndices.from_ranges(),)  # as in seabreeze-3.0.9
    integration_time_min = 1000
    integration_time_max = 1600000000
    integration_time_base = 1000
    spectrum_num_pixel = 512
    spectrum_raw_length = (512 * 2) + 1
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureNIRQUEST512,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class NIRQUEST256(SeaBreezeDevice):

    model_name = "NIRQUEST256"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1028
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = (DarkPixelIndices.from_ranges(),)  # as in seabreeze-3.0.9
    integration_time_min = 1000
    integration_time_max = 1600000000
    integration_time_base = 1000
    spectrum_num_pixel = 256
    spectrum_raw_length = (256 * 2) + 1
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureNIRQUEST256,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class MAYA2000PRO(SeaBreezeDevice):

    model_name = "MAYA2000PRO"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x102A
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((0, 4), (2064, 2068))
    integration_time_min = 7200
    integration_time_max = 65000000
    integration_time_base = 1
    spectrum_num_pixel = 2304
    spectrum_raw_length = (2304 * 2) + 1
    spectrum_max_value = 64000
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureMAYA2000PRO,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class MAYA2000(SeaBreezeDevice):

    model_name = "MAYA2000"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x102C
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((0, 8), (2072, 2080))
    integration_time_min = 15000
    integration_time_max = 1600000000
    integration_time_base = 1
    spectrum_num_pixel = 2304
    spectrum_raw_length = (2304 * 2) + 1
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported("NORMAL", "SOFTWARE", "HARDWARE")

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureMAYA2000,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class TORUS(SeaBreezeDevice):

    model_name = "TORUS"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1040
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges()
    integration_time_min = 1000
    integration_time_max = 655350000
    integration_time_base = 1
    spectrum_num_pixel = 2048
    spectrum_raw_length = (2048 * 2) + 1
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureTORUS,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class APEX(SeaBreezeDevice):

    model_name = "APEX"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1044
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((0, 4), (2064, 2068))
    integration_time_min = 15000
    integration_time_max = 1600000000
    integration_time_base = 1
    spectrum_num_pixel = 2304
    spectrum_raw_length = (2304 * 2) + 1
    spectrum_max_value = 64000
    trigger_modes = TriggerMode.supported(
        "NORMAL",
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureAPEX,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class MAYALSL(SeaBreezeDevice):

    model_name = "MAYALSL"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1046
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((0, 4), (2064, 2068))
    integration_time_min = 7200
    integration_time_max = 65000000
    integration_time_base = 1
    spectrum_num_pixel = 2304
    spectrum_raw_length = (2304 * 2) + 1
    spectrum_max_value = 64000
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureMAYALSL,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class JAZ(SeaBreezeDevice):

    model_name = "JAZ"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x2000
    usb_endpoint_map = EndPointMap(ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82)
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((2, 24))
    integration_time_min = 1000
    integration_time_max = 655350000
    integration_time_base = 1
    spectrum_num_pixel = 2048
    spectrum_raw_length = 2048 * 2  # XXX: No Sync byte!
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureJAZ,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class STS(SeaBreezeDevice):

    model_name = "STS"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x4000
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81
    )  # XXX: we'll ignore the alternative EPs
    usb_protocol = OBPProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges()
    integration_time_min = 10
    integration_time_max = 85000000
    integration_time_base = 1
    spectrum_num_pixel = 1024
    spectrum_raw_length = 1024 * 2
    spectrum_max_value = 16383
    trigger_modes = TriggerMode.supported("OBP_NORMAL", "OBP_EXTERNAL", "OBP_INTERNAL")

    # features
    feature_classes = (
        sbf.spectrometer.SeaBreezeSpectrometerFeatureSTS,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class QEPRO(SeaBreezeDevice):

    model_name = "QEPRO"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x4004
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81
    )  # XXX: we'll ignore the alternative EPs
    usb_protocol = OBPProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((0, 4), (1040, 1044))
    integration_time_min = 8000
    integration_time_max = 1600000000
    integration_time_base = 1
    spectrum_num_pixel = 1044
    spectrum_raw_length = (1044 * 4) + 32  # XXX: Metadata
    spectrum_max_value = (2**18) - 1
    trigger_modes = TriggerMode.supported("NORMAL", "LEVEL", "SYNCHRONIZATION", "EDGE")

    # features
    feature_classes = (
        sbf.spectrometer.SeaBreezeSpectrometerFeatureQEPRO,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class VENTANA(SeaBreezeDevice):

    model_name = "VENTANA"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x5000
    usb_endpoint_map = EndPointMap(ep_out=0x01, lowspeed_in=0x82)
    usb_protocol = OBPProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges()
    integration_time_min = 22000
    integration_time_max = 60000000
    integration_time_base = 1
    spectrum_num_pixel = 1024
    spectrum_raw_length = 1024 * 2  # XXX: No Sync byte!
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported("NORMAL", "LEVEL", "SYNCHRONIZATION", "EDGE")

    # features
    feature_classes = (
        sbf.spectrometer.SeaBreezeSpectrometerFeatureVENTANA,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class SPARK(SeaBreezeDevice):

    model_name = "SPARK"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x4200
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81
    )  # XXX: we'll ignore the alternative EPs
    usb_protocol = OBPProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges()
    integration_time_min = 10
    integration_time_max = 85000000
    integration_time_base = 1
    spectrum_num_pixel = 1024
    spectrum_raw_length = (1024 * 2) + 64  # XXX: Metadata
    spectrum_max_value = 16383
    trigger_modes = TriggerMode.supported("OBP_NORMAL", "OBP_EXTERNAL", "OBP_INTERNAL")

    # features
    feature_classes = (
        sbf.spectrometer.SeaBreezeSpectrometerFeatureSPARK,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class HDX(SeaBreezeDevice):

    model_name = "HDX"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x2003
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OBPProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges()
    integration_time_min = 6000
    integration_time_max = 10000000
    integration_time_base = 1
    spectrum_num_pixel = 2068
    spectrum_raw_length = (2068 * 2) + 64  # XXX: Metadata
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported(
        "OBP_NORMAL", "OBP_LEVEL", "OBP_EDGE", "DISABLED"
    )

    # features
    feature_classes = (
        sbf.spectrometer.SeaBreezeSpectrometerFeatureHDX,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
        sbf.nonlinearity.NonlinearityCoefficientsFeatureOBP,
    )



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeIPv4Feature(SeaBreezeFeature):
    identifier = "ipv4"

    def get_dhcp_enable_state(self, interface_index: int) -> bool:
        raise NotImplementedError("implement in derived class")

    def set_dhcp_enable_state(self, interface_index: int, is_enabled: bool) -> None:
        raise NotImplementedError("implement in derived class")

    def get_number_of_ipv4_addresses(self, interface_index: int) -> int:
        raise NotImplementedError("implement in derived class")

    def get_ipv4_address(self, interface_index: int, address_index: int) -> str:
        raise NotImplementedError("implement in derived class")

    def get_default_gateway(self, interface_index: int) -> str:
        raise NotImplementedError("implement in derived class")

    def set_default_gateway(
        self, interface_index: int, default_gateway_address: str
    ) -> None:
        raise NotImplementedError("implement in derived class")

    def add_ipv4_address(self, interface_index: int, ipv4_address: str) -> None:
        raise NotImplementedError("implement in derived class")

    def delete_ipv4_address(self, interface_index: int, address_index: int) -> None:
        raise NotImplementedError("implement in derived class")



"""This is a wrapper for the pyusb-implementation of the seabreeze-library

"""
from seabreeze.pyseabreeze.api import SeaBreezeAPI
from seabreeze.pyseabreeze.devices import SeaBreezeDevice
from seabreeze.pyseabreeze.exceptions import SeaBreezeError
from seabreeze.pyseabreeze.exceptions import SeaBreezeNumFeaturesError

# noinspection PyUnresolvedReferences
from seabreeze.pyseabreeze.features import SeaBreezeAcquisitionDelayFeature
from seabreeze.pyseabreeze.features import SeaBreezeContinuousStrobeFeature
from seabreeze.pyseabreeze.features import SeaBreezeDataBufferFeature
from seabreeze.pyseabreeze.features import SeaBreezeDHCPServerFeature
from seabreeze.pyseabreeze.features import SeaBreezeEEPROMFeature
from seabreeze.pyseabreeze.features import SeaBreezeEthernetConfigurationFeature
from seabreeze.pyseabreeze.features import SeaBreezeFastBufferFeature
from seabreeze.pyseabreeze.features import SeaBreezeGPIOFeature
from seabreeze.pyseabreeze.features import SeaBreezeI2CMasterFeature
from seabreeze.pyseabreeze.features import SeaBreezeIntrospectionFeature
from seabreeze.pyseabreeze.features import SeaBreezeIPv4Feature
from seabreeze.pyseabreeze.features import SeaBreezeIrradCalFeature
from seabreeze.pyseabreeze.features import SeaBreezeLightSourceFeature
from seabreeze.pyseabreeze.features import SeaBreezeMulticastFeature
from seabreeze.pyseabreeze.features import SeaBreezeNetworkConfigurationFeature
from seabreeze.pyseabreeze.features import SeaBreezeNonlinearityCoefficientsFeature
from seabreeze.pyseabreeze.features import SeaBreezeOpticalBenchFeature
from seabreeze.pyseabreeze.features import SeaBreezePixelBinningFeature
from seabreeze.pyseabreeze.features import SeaBreezeRawUSBBusAccessFeature
from seabreeze.pyseabreeze.features import SeaBreezeRevisionFeature
from seabreeze.pyseabreeze.features import SeaBreezeShutterFeature
from seabreeze.pyseabreeze.features import SeaBreezeSpectrometerFeature
from seabreeze.pyseabreeze.features import SeaBreezeSpectrumProcessingFeature
from seabreeze.pyseabreeze.features import SeaBreezeStrayLightCoefficientsFeature
from seabreeze.pyseabreeze.features import SeaBreezeStrobeLampFeature
from seabreeze.pyseabreeze.features import SeaBreezeTemperatureFeature
from seabreeze.pyseabreeze.features import SeaBreezeThermoElectricFeature
from seabreeze.pyseabreeze.features import SeaBreezeWifiConfigurationFeature

# noinspection PyUnresolvedReferences,PyProtectedMember
from seabreeze.pyseabreeze.features._base import SeaBreezeFeature

__all__ = [
    "_backend_",
    "SeaBreezeAcquisitionDelayFeature",
    "SeaBreezeAPI",
    "SeaBreezeContinuousStrobeFeature",
    "SeaBreezeDataBufferFeature",
    "SeaBreezeDevice",
    "SeaBreezeDHCPServerFeature",
    "SeaBreezeEEPROMFeature",
    "SeaBreezeError",
    "SeaBreezeEthernetConfigurationFeature",
    "SeaBreezeFastBufferFeature",
    "SeaBreezeFeature",
    "SeaBreezeGPIOFeature",
    "SeaBreezeI2CMasterFeature",
    "SeaBreezeIntrospectionFeature",
    "SeaBreezeIPv4Feature",
    "SeaBreezeIrradCalFeature",
    "SeaBreezeLightSourceFeature",
    "SeaBreezeMulticastFeature",
    "SeaBreezeNetworkConfigurationFeature",
    "SeaBreezeNonlinearityCoefficientsFeature",
    "SeaBreezeNumFeaturesError",
    "SeaBreezeOpticalBenchFeature",
    "SeaBreezePixelBinningFeature",
    "SeaBreezeRawUSBBusAccessFeature",
    "SeaBreezeRevisionFeature",
    "SeaBreezeShutterFeature",
    "SeaBreezeSpectrometerFeature",
    "SeaBreezeSpectrumProcessingFeature",
    "SeaBreezeStrayLightCoefficientsFeature",
    "SeaBreezeStrobeLampFeature",
    "SeaBreezeTemperatureFeature",
    "SeaBreezeThermoElectricFeature",
    "SeaBreezeWifiConfigurationFeature",
]

_backend_ = "pyseabreeze"



"""Implementation of the Seabreeze Transport layer.

Some spectrometers can support different transports (usb, network, rs232, etc.)

"""
from __future__ import annotations

import importlib
import inspect
import logging
import warnings
from functools import partialmethod
from typing import TYPE_CHECKING
from typing import Any
from typing import Iterable
from typing import Tuple

import usb.backend
import usb.core
import usb.util

from seabreeze.pyseabreeze.types import PySeaBreezeProtocol
from seabreeze.pyseabreeze.types import PySeaBreezeTransport

if TYPE_CHECKING:
    from seabreeze.pyseabreeze.devices import EndPointMap


# encapsulate usb.core.USBError
class USBTransportError(Exception):
    def __init__(
        self, *args: Any, errno: int | None = None, error_code: int | None = None
    ) -> None:
        super().__init__(*args)
        self.errno = errno
        self.backend_error_code = error_code

    @classmethod
    def from_usberror(cls, err: usb.core.USBError) -> USBTransportError:
        return cls(str(err), errno=err.errno, error_code=err.backend_error_code)


class USBTransportDeviceInUse(Exception):
    pass


DeviceIdentity = Tuple[int, int, int, int]


# this can and should be opaque to pyseabreeze
class USBTransportHandle:
    def __init__(self, pyusb_device: usb.core.Device) -> None:
        """encapsulation for pyusb device classes

        Parameters
        ----------
        pyusb_device
        """
        self.pyusb_device: usb.core.Device = pyusb_device
        # noinspection PyUnresolvedReferences
        self.identity: DeviceIdentity = (
            pyusb_device.idVendor,
            pyusb_device.idProduct,
            pyusb_device.bus,
            pyusb_device.address,
        )
        self.pyusb_backend = get_name_from_pyusb_backend(pyusb_device.backend)

    def close(self) -> None:
        try:
            self.pyusb_device.reset()
        except usb.core.USBError:
            logging.debug(
                "USBError while calling USBTransportHandle.close on {:04x}:{:04x}".format(
                    self.identity[0], self.identity[1]
                ),
                exc_info=True,
            )

    def __del__(self) -> None:
        if self.pyusb_backend == "libusb1":
            # have to check if .finalize() has been called
            # -> todo: maybe better to fix this in the api initialization of cseabreeze
            # -> todo: will probably have to check pyusb versions and only do this when necessary
            if not getattr(self.pyusb_device.backend, "_finalize_called", False):
                # if usb.core.Device.reset() gets called but the backend has been finalized already
                # (this happens only during interpreter shutdown)
                self.close()
        else:
            self.close()
        self.pyusb_device = None


class USBTransport(PySeaBreezeTransport[USBTransportHandle]):
    """implementation of the usb transport interface for spectrometers"""

    _required_init_kwargs = ("usb_product_id", "usb_endpoint_map", "usb_protocol")
    vendor_id = 0x2457
    product_ids: dict[int, str] = {}

    # add logging
    _log = logging.getLogger(__name__)

    def __init__(
        self,
        usb_product_id: int,
        usb_endpoint_map: EndPointMap,
        usb_protocol: type[PySeaBreezeProtocol],
    ) -> None:
        super().__init__()
        self._product_id = usb_product_id
        self._endpoint_map = usb_endpoint_map
        self._protocol_cls = usb_protocol
        # internal settings
        self._default_read_size = {
            "low_speed": 64,
            "high_speed": 512,
            "high_speed_alt": 512,
        }
        self._read_endpoints = {
            "low_speed": "lowspeed_in",
            "high_speed": "highspeed_in",
            "high_speed_alt": "highspeed_in2",
        }
        self._default_read_endpoint = "low_speed"
        self._default_read_spectrum_endpoint = "high_speed"
        # internal state
        self._device: USBTransportHandle | None = None
        self._opened: bool | None = None
        self._protocol: PySeaBreezeProtocol | None = None

    def open_device(self, device: USBTransportHandle) -> None:
        if not isinstance(device, USBTransportHandle):
            raise TypeError("device needs to be a USBTransportHandle")
        # device.reset()
        self._device = device
        pyusb_device = self._device.pyusb_device
        try:
            if pyusb_device.is_kernel_driver_active(0):
                pyusb_device.detach_kernel_driver(0)
        except NotImplementedError:
            pass  # unavailable on some systems/backends
        try:
            pyusb_device.set_configuration()
        except usb.core.USBError as err:
            if err.errno == 16:
                # TODO: warn as in cseabreeze
                self._opened = True
                raise USBTransportDeviceInUse(
                    "device probably used by another thread/process"
                )
            raise USBTransportError.from_usberror(err)
        else:
            self._opened = True
        # This will initialize the communication protocol
        if self._opened:
            self._protocol = self._protocol_cls(self)

    @property
    def is_open(self) -> bool:
        return self._opened or False

    def close_device(self) -> None:
        if self._device is not None:
            self._device.close()
            self._device = None
        self._opened = False
        self._protocol = None

    def write(self, data: bytes, timeout_ms: int | None = None, **kwargs: Any) -> int:
        if self._device is None:
            raise RuntimeError("device not opened")
        if kwargs:
            warnings.warn(f"kwargs provided but ignored: {kwargs}")
        return self._device.pyusb_device.write(  # type: ignore
            self._endpoint_map.ep_out, data, timeout=timeout_ms
        )

    def read(
        self,
        size: int | None = None,
        timeout_ms: int | None = None,
        mode: str | None = None,
        **kwargs: Any,
    ) -> bytes:
        if self._device is None:
            raise RuntimeError("device not opened")
        mode = mode if mode is not None else self._default_read_endpoint
        endpoint = getattr(self._endpoint_map, self._read_endpoints[mode])
        if size is None:
            size = self._default_read_size[mode]
        if kwargs:
            warnings.warn(f"kwargs provided but ignored: {kwargs}")
        ret: bytes = self._device.pyusb_device.read(
            endpoint, size, timeout=timeout_ms
        ).tobytes()
        return ret

    @property
    def default_timeout_ms(self) -> int:
        if not self._device:
            raise RuntimeError("no protocol instance available")
        return self._device.pyusb_device.default_timeout  # type: ignore

    @property
    def protocol(self) -> PySeaBreezeProtocol:
        if self._protocol is None:
            raise RuntimeError("no protocol instance available")
        return self._protocol

    @classmethod
    def list_devices(cls, **kwargs: Any) -> Iterable[USBTransportHandle]:
        """list pyusb devices for all available spectrometers

        Note: this includes spectrometers that are currently opened in other
        processes on the machine.

        Yields
        ------
        devices : USBTransportHandle
            unique pyusb devices for each available spectrometer
        """
        # check if a specific pyusb backend is requested
        _pyusb_backend = kwargs.get("pyusb_backend", None)
        # get all matching devices
        try:
            pyusb_devices = usb.core.find(
                find_all=True,
                custom_match=lambda dev: (
                    dev.idVendor == cls.vendor_id and dev.idProduct in cls.product_ids
                ),
                backend=get_pyusb_backend_from_name(name=_pyusb_backend),
            )
        except usb.core.NoBackendError:
            raise RuntimeError("No pyusb backend found")
        # encapsulate
        for pyusb_device in pyusb_devices:
            yield USBTransportHandle(pyusb_device)

    @classmethod
    def register_model(cls, model_name: str, **kwargs: Any) -> None:
        product_id = kwargs.get("usb_product_id")
        if not isinstance(product_id, int):
            raise TypeError(f"product_id {product_id:r} not an integer")
        if product_id in cls.product_ids:
            raise ValueError(f"product_id 0x{product_id:04x} already in registry")
        cls.product_ids[product_id] = model_name

    @classmethod
    def supported_model(cls, device: USBTransportHandle) -> str | None:
        """return supported model

        Parameters
        ----------
        device : USBTransportHandle
        """
        if not isinstance(device, USBTransportHandle):
            return None
        # noinspection PyUnresolvedReferences
        return cls.product_ids[device.pyusb_device.idProduct]

    @classmethod
    def specialize(cls, model_name: str, **kwargs: Any) -> type[USBTransport]:
        assert set(kwargs) == set(cls._required_init_kwargs)
        # usb transport register automatically on registration
        cls.register_model(model_name, **kwargs)
        specialized_class = type(
            f"USBTransport{model_name}",
            (cls,),
            {"__init__": partialmethod(cls.__init__, **kwargs)},
        )
        return specialized_class

    @classmethod
    def initialize(cls, **_kwargs: Any) -> None:
        for device in cls.list_devices(**_kwargs):
            try:
                device.pyusb_device.reset()
                # usb.util.dispose_resources(device)  <- already done by device.reset()
            except Exception as err:
                cls._log.debug(
                    "initialize failed: {}('{}')".format(
                        err.__class__.__name__, getattr(err, "message", "no message")
                    )
                )

    @classmethod
    def shutdown(cls, **_kwargs: Any) -> None:
        # dispose usb resources
        for device in cls.list_devices(**_kwargs):
            try:
                usb.util.dispose_resources(device.pyusb_device)
            except Exception as err:
                cls._log.debug(
                    "shutdown failed: {}('{}')".format(
                        err.__class__.__name__, getattr(err, "message", "no message")
                    )
                )


_pyusb_backend_instances: dict[str, usb.backend.IBackend] = {}


def get_pyusb_backend_from_name(name: str) -> usb.backend.IBackend:
    """internal: allow requesting a specific pyusb backend for testing"""
    if name is None:
        # default is pick first that works: ('libusb1', 'libusb0', 'openusb')
        _backend = None
    else:
        try:
            _backend = _pyusb_backend_instances[name]
        except KeyError:
            try:
                m = importlib.import_module(f"usb.backend.{name}")
            except ImportError:
                raise RuntimeError(f"unknown pyusb backend: {name!r}")
            # noinspection PyUnresolvedReferences
            _backend = m.get_backend()
            # raise if a pyusb backend was requested but can't be loaded
            if _backend is None:
                raise RuntimeError(f"pyusb backend failed to load: {name!r}")
            _pyusb_backend_instances[name] = _backend
    return _backend


def get_name_from_pyusb_backend(backend: usb.backend.IBackend) -> str | None:
    """internal: return backend name from loaded backend"""
    module = inspect.getmodule(backend)
    if not module:
        return None
    return module.__name__.split(".")[-1]



import logging
import sys
import warnings
from typing import Any
from typing import Dict
from typing import Optional
from typing import cast

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from seabreeze.types import SeaBreezeBackend

__all__ = ["use", "get_backend"]


class BackendConfig:
    """*internal* config dict"""

    requested = "cseabreeze"  # default is cseabreeze
    available = ("cseabreeze", "pyseabreeze")
    allow_fallback = False
    api_kwargs: Dict[str, Any] = {}  # for pytests


def use(
    backend: Literal["cseabreeze", "pyseabreeze"], force: bool = True, **kwargs: Any
) -> None:
    """
    select the backend used for communicating with the spectrometer

    Parameters
    ----------
    backend : str

    force : bool, optional, default: True
        raises an ImportError when ``seabreeze.get_backend()`` is called
        and the requested backend can not be imported. force=True should
        be used in user code to ensure that the correct backend is being
        loaded.

    pyseabreeze only
    ----------------
    pyusb_backend: str
        either libusb1, libusb0 or openusb

    """
    if backend not in BackendConfig.available:
        raise ValueError(
            "backend not in: {}".format(", ".join(BackendConfig.available))
        )
    if "seabreeze.spectrometers" in sys.modules:
        warnings.warn(
            "seabreeze.use has to be called before importing seabreeze.spectrometers",
            stacklevel=2,
        )

    BackendConfig.requested = backend
    BackendConfig.allow_fallback = not force
    BackendConfig.api_kwargs = {}

    if backend == "pyseabreeze":
        if "pyusb_backend" in kwargs:
            pyusb_backend = kwargs.pop("pyusb_backend")
            BackendConfig.api_kwargs["pyusb_backend"] = pyusb_backend
    if kwargs:
        raise TypeError(
            f"unknown keyword arguments {set(kwargs)!r} for backend {backend!r}"
        )


def get_backend() -> SeaBreezeBackend:
    """
    return the requested backend or a fallback. configuration is done
    via ``seabreeze.use()``

    Returns
    -------
    backend:
        a backend interface for communicating with the spectrometers
    """

    def _use_backend(name: str) -> Optional[SeaBreezeBackend]:
        # internal: import the libseabreeze cython wrapper -> cseabreeze
        try:
            if name == "cseabreeze":
                import seabreeze.cseabreeze as sbb
            elif name == "pyseabreeze":
                import seabreeze.pyseabreeze as sbb  # type: ignore
            else:
                raise ValueError(f"unknown backend {name!r}")
        except ImportError as err:
            logging.getLogger(__name__).warning(
                f"seabreeze can't load {name!r} backend - error: '{err!r}'",
                exc_info=True,
            )
            return None
        else:
            return cast(SeaBreezeBackend, sbb)

    requested = BackendConfig.requested
    fallback = BackendConfig.allow_fallback
    backends = BackendConfig.available

    backend = _use_backend(requested)  # trying to import requested backend
    if backend is None and fallback:

        warnings.warn(
            f"seabreeze failed to load user requested {requested!r} backend but will try fallback backends",
            category=ImportWarning,
        )

        fallback_backends = [b for b in backends if b != requested]
        for fallback_backend in fallback_backends:
            backend = _use_backend(fallback_backend)
            if backend is not None:
                break
        else:
            raise ImportError(
                "Could not import backend. Tried: {}".format(", ".join(backends))
            )

    if backend is None:
        raise ImportError(f"Could not import backend. Requested: {requested}")

    # provide for testing purposes
    backend._api_kwargs = BackendConfig.api_kwargs

    return backend
